// markgs_X4f5c.C as markgs_X4f5b.C - but just minor
// speedup in cut search - discard L-cuts (and extensions) where
// 2log_2(L!) exceeds quality bound

/*

plantri -a -d 19 -c4 | head | perl -f asc2math2x.prl2 | gawk -f markg/dup_vertex_1.awk > /tmp/KK

cat -n /tmp/KK | markg/markgs_X4f5b

*/
#define MAX_CUT_SIZE 9
#define MAX_UNIVAL_SIZE 12


#define c_int const int
#define integer int
#define operation oper

#include"mydefs.h"
#define POLY2
// to get alternative polynomial output easier for MATH
#include"poly.h"
// cnts non-iso max knot geners

// #define ERROR(u) if(u) { cerr<<"error line " << __LINE__ << endl; }

// int X,X1,nr,nr_aut,MAX_MARKS=30000;
// X = # of verts
// X1 = max poss deg of W_+- = 2+X1/4
// nr = graph id

template<class C>
class array {
 public:
  int len;
  C *els;

  ~array() { delete [] els; }
  array(): len(0), els(0) {}
  array(int l): len(l), els(new C[l]) {}
  array(int l, const C & o): len(l), els(new C[l])
    { for(int i=0;i<len;++i) els[i]=o; }
  array(C *p, int l): len(l), els(copyarray(p,l)) {}
  array(const array &u): len(u.len), els(copyarray(u.els,u.len)) {}
  C & operator[] (int x) { ret els[x]; }
  const array & operator= (const array &u) 
  { delete[] els; len=u.len; els=copyarray(u.els,u.len); ret u; }
  operator C * () const { ret els; }
};

typedef array<int> onear;
typedef array<onear> twoar;
typedef array<twoar> trear;

class graph {
  public:
    int X, Y, *edges;
    // an unitriv graph with X 3val Yunval verts
    // edges is int[3X] and has the 3 adj verts
    // to each 3val vertex clockw
    // unival verts are #ed by negative numbers
    //  -1   -2
    //  \    /
    //   \  /
    //    \/      X=2
    //    | 1     Y=4
    //    |	  edges={2,-2,-1,  /* vert1 */  -3,-4,1  /* vert2 */}
    //    | 2
    //    /\
    //   /  \
    //  /    \
    //  -3  -4 

  graph(c_int x=0,c_int y=0): X(x), Y(y), edges(new int[3*X]) {}
  graph(c_int x,c_int y, c_int *t): X(x), Y(y),
    edges(copyarray(t,3*x)) {}
  graph(const graph & g): X(g.X), Y(g.Y),
    edges(copyarray(g.edges,3*g.X)) {}
  graph & oper = (const graph & g) {
    delete [] edges;
    X=g.X;
    Y=g.Y;
    edges=copyarray(g.edges,3*X);
    ret *this;
  }
  friend istream & oper >>(istream &i, graph & g)
  {
    i>>g.X>>g.Y; int nd=3*g.X;
    g.edges=new int[nd];
    for(int j=0;j<nd;++j) { i>>g[j]; if(g[j]>=0) --g[j]; }
    ret i;
  }
  friend ostream & oper << (ostream & o, const graph & g)
  {
    cout<<"["<<g.X<<","<<g.Y<<"] ";
    for(int j=0;j<g.X;++j)
      varoutputarray(cout,g.edges+3*j,3,"{",",","}");
    ret cout<<endl;
  }
  int & oper [] (c_int i) { ret edges[i]; }
};

inline int fac(int n, bool log_flag=0) {
// with log flag= true it returns rounded (log_2(n!^2))
// which is neede for complexity estimate
  static integer db[MAX_CUT_SIZE+1], mv=-1;
  static long double dbl[MAX_CUT_SIZE+1], log2=2./logl(2.);
  ERROR(n<0 || n>MAX_CUT_SIZE);
  if(n<=mv) ret log_flag ? (int)roundl(dbl[n]) : db[n];

  for(integer j=mv+1;j<=n;++j) {
    db[j]= j<2 ? 1  : j*db[j-1];
    dbl[j]=j<2 ? 0. : dbl[j-1]+log2*logl((long double)j);
  }
  mv=n;
  ret log_flag ? (int)roundl(dbl[n]) : db[n];
}

// the two routines that change a perm of Y els into a number between
// 0 and Y!-1 and back

inline int code(c_int *perm, int Y)
{
  int g[Y],i,j,h=fac(Y),r=0;
  for(j=0;j<Y;++j) g[j]=j+1;
  for(j=0;j<Y;++j)
  {
    h/=(Y-j);
    for(i=0;i<Y && g[i]!=perm[j];++i) if(g[i]) r+=h;
    ERROR(i==Y);
    // TRCA(g,Y);
    // TRC(r);
    g[i]=0;
  }
  ERROR(r>fac(Y));
  ret r;
}

inline int *decode(int *perm, int r, c_int Y)
{
  // TRC(Y);
  // TRC(r);
  int g[Y],i,j,h=fac(Y);
  for(j=0;j<Y;++j) g[j]=j+1;
  for(j=0;j<Y;++j)
  {
    h/=(Y-j);
    for(i=0;i<Y; ++i) if(g[i]) { if(r<h) break; r-=h; }
    ERROR(i==Y);
    perm[j]=g[i];
    g[i]=0;
  }
  ERROR(r);
  // TRCA(perm,Y);
  ret perm;
}

inline int *inverse(int *perm, int *inve, c_int r)
{
  for(int i=0;i<r;++i) inve[perm[i]-1]=i+1;
  ret inve;
}

inline int *inverse_table(int *it, c_int r)
{
  int h=fac(r), perm[r], inve[r];
  while(h--) {
    // TRC(h);
    decode(perm,h,r);
    // TRCA(perm,r);
    inverse(perm,inve,r);
    // TRCA(inve,r);
    it[h]=code(inve,r);
    // TRC(it[h]);
  } 
  ret it;
}

#define TWOAR(a,d) twoar a(d,onear(d,-1)); 


inline int combine(c_int *p1, c_int *p2, c_int Y)
{
  // TRCA(p1,Y);
  // TRCA(p2,Y);
  // counts cycles of p1=circ p2
  int p, k, l, cyc[Y];
  zeroarray(cyc,Y,-1);
  for(k=p=0;k<Y;++k)
  if(cyc[k]<0) { ++p; l=k; 
    do { cyc[l]=p; l=p2[p1[l]-1]-1; } while (l!=k);
  }
  // TRC(p);
  ret p;
}

poly1var combine(poly1var *A, poly1var *B, int Y)
{
  // combines the local decompositions of two 13-gr giving
  // a 3val gr

  poly1var c(0);
  int i,j, p1[Y], p2[Y], yf=fac(Y);
  for(i=0;i<yf;++i) if(A[i])
  for(j=0;j<yf;++j) if(B[j])
  c+=(A[i]*B[j]>>combine(decode(p1,i,Y),decode(p2,j,Y),Y));
  ret c;
}

inline int partial_combine(c_int *p1, c_int *p2,
  c_int Y1, c_int Y2, c_int N, int *p, c_int *TT, c_int *UR)
{
  // TRCA(p1,Y1);
  // TRCA(p2,Y2);
  // TRCA(TT,Y1-N);
  // TRCA(UR,Y2-N);
  // combines the first N strings pf perms p1 in Yi
  // stirngs into a perm of Y1+Y2-2n strings written in p
  // the strings in p are nred 1...Y1-n= n+1...Y1 of p1
  // Y1+1...Y1+Y2-2n=n+1...Y2 of p2
  // returned in the number of closee cycles
  int cyc[Y1], q, r, i, x;
  zeroarray(cyc,Y1);

  for(i=N;i<Y1;++i)
  {
    q=1; x=i;
    do {
#define DO1 \
      if(q==1) { cyc[x]=1; x=p1[x]-1; } else x=p2[x]-1; q=3-q;
      DO1
    } while (x<N);
    // q was switched wrongly by q=3-q; before loop broken out
    p[TT[i-N]-1]= q==2? TT[x-N] : UR[x-N];
    // TRC(TT[i-N]);
    // TRC(p[TT[i-N]-1]);
  }
  for(i=N;i<Y2;++i)
  {
    q=2; x=i;
    do {
      DO1
    } while (x<N);
    p[UR[i-N]-1]= q==2? TT[x-N] : UR[x-N];
    // TRC(UR[i-N]);
    // TRC(p[UR[i-N]-1]);
  }
  for(i=r=0;i<N;++i)
  if(cyc[i]==0)
  { // TRC(i);
    ++r; q=1; x=i;
    do {
      DO1
    } while (x-i || q-1);
    // TRC(i);
  }
  // TRCA(p,Y1+Y2-2*N);
  ret r;
}

void partial_combine(const poly1var *A, const poly1var *B, c_int Y1,
  c_int Y2, c_int N, poly1var *C, c_int *TT, c_int *UR)
{
  // partialy combines the local decompositions of two 13-gr giving
  // a 13val gr; C[] should have size (Y1+Y2-2*N)!

  poly1var c;
  int Y3=Y1+Y2-2*N, i,j, p1[Y1], p2[Y2], p[Y3],
    y1f=fac(Y1), y2f=fac(Y2);
  zeroarray(C,fac(Y3));

  for(i=0;i<y1f;++i) if(A[i])
  { decode(p1,i,Y1);
    // TRC(i);
    // TRCA(p1,Y1);
    for(j=0;j<y2f;++j) if(B[j])
    {
      // TRC(j);
      c=A[i]*B[j]>>partial_combine(p1,decode(p2,j,Y2),Y1,Y2,N,p,TT,UR);
      // TRC(j);
      // TRCA(p,Y3);
      C[code(p,Y3)]+=c;
      // TRCA(p,Y3);
    }
  }
  // TRC(1);
}

// next the routines for finding a good 4- or 5-cut

template<class C>
inline C max(const C *a, const int l) 
{ int i=0;
  C r=a[0];
  for(;i<l;++i) if(a[i]>r) r=a[i];
  ret r;
}

int cut_quality(const graph & g, c_int *cut, c_int L, c_int bound,
  c_int *s, c_int *e, int *parity)
{

  // estimates the complexity of computation via this
  // cut and writes into parity 
  // the lower the score >0 the better
  // ret value 0 - cut and any extension of it is not good
  // -1 it is not a cut (can try to extend)

  ERROR(L<0);
  if(L==0) ret -1;
  // THIS SHOULD INCLUDE CUT TEST!!!

  // for a cut of L edges in cut[] gives the smaller between the
  // number of vertices on either side of the cut;
  // the arrays s and e are determined from g and are passed 
  // as argument from four_cut() to save computation
  // parity[]=integer[X] is output param and parity[i] records
  // which part (1 or 2) vertex i belongs to after the cut

  // if cut[L] is not a cut, return 0,-1;
  // otherwise retun (L!)^2+2^ops[1]+2^ops[2]

  // HERE parity[] MUST BE ACCESIBLE ALSO WITH INDICES -1 to -L
  // of unival vertices

  // ERROR(g.Y); should work also for uni3val graphs
  int X=g.X, Y=g.Y, p, np, r=X+Y-2, o, i, j, l,
    ops[3]={fac(L,true),1,1}, pps[3]={0,L,L}, ne=(3*X+Y)/2;
    // ops[1 & 2] count the number of 3val verts in part 1 & 2
    // pps[1 & 2] count the number of 1val verts in part 1 & 2
    // count already 1 vertex by below initialization
    // r = number of non-assigned vertices
  if(ops[0]>=bound) ret 0;

  zeroarray(parity-Y,X+Y);

  for(i=0;i<L;++i) 
    if(parity[s[cut[i]]]++ || parity[e[cut[i]]]++) ret 0;
    // must explicitly avoid cuts in which two edges
    // incident from same vertex occur
    // they are useless anyway

  zeroarray(parity-Y,X+Y);
  // const integer test[4]={1,13,16,25};
  // TRCA(cut,L);
  // bool TEST=(L==4 && diffarray(cut,test,L)==0);
  // TEST=true;
  
  /*
  if(TEST) {
    TRCA(cut,L);
    TRC(cut[0]);
    TRCA(s,(3*X+Y)/2);
    TRCA(e,(3*X+Y)/2);
    TRC(ss[0]);
    TRC(se[0]);
  }
  */
  

  parity[s[cut[0]]]=1;
  parity[e[cut[0]]]=2;
  // TRC(g);
  // if(TEST) TRC(ss[0]);
  // if(TEST) TRC(se[0]);
  // TRC(Y);
  // if(TEST) TRCA(parity-Y,X+Y);

  // TRC(g);
  // TRCA(cut,L);
  // TRCA(ss,L);
  // TRCA(se,L);
  integer cons_flag=1, status[ne];
  // consitency flag; after r is set down to 0
  // run 1 more time to check consitency!

  for(i=0;i<ne;++i)  status[i]=1;
  for(i=0;i<L;++i)  status[cut[i]]=2;

  do {
    // TRCA(parity,X);
    // TRC(r);
    for(i=0;i<ne;++i) 
    { integer &st=status[i];
    
      if(st>0) {
        integer w[2]={s[i],e[i]},h;
        for(h=0;h<2;++h)
        if(p=parity[w[h]])
        { 
          np= st==2 ? 3-p : p;
          ERROR(p-1 && p-2)
          if(parity[w[1-h]])
	  {
	     if(parity[w[1-h]]!=np) ret -1;
	     st=0;
	     break;
	  }
	  parity[w[1-h]]=np;
          // TRC(Y);
          // if(TEST) TRCA(parity-Y,X+Y);
          // TRC(i);
          // if(TEST) TRC(l);
          // if(TEST) TRC(p);
	  st=0;
          if(w[1-h]>=0) ++ops[np];
          // count only trivalent vertices for cut quality
          else if(++pps[np]>MAX_UNIVAL_SIZE) ret 0;
          // do not wat more than 7 unival vertices per graph
          // TRC(r);
          --r;
          // TRC(r);
          // TRCA(parity-Y,Y+X);
          ERROR(r<0)
          // if(r<0) { TRC(X); TRC(Y); TRCA(parity-Y,Y+X); TRC(r); }
	  break;
        }
      }
    }
  } while(r || cons_flag--);
  // TRCA(parity-Y,Y+X);
  // TRCA(cut,L);
  // TRC(ops[1]);
  // TRC(ops[2]);
  // TRC(L);
  // TRC(fac(L));
  // TRC( 1<<(ops[1]-1)+1<<(ops[2]-1)+fac(L)*fac(L) );
  // we used he sofar dummied ops[0] for computing the 
  // expeced complexity of the cut
  ret max(ops,3);
}

int look_for_cut(const graph & g, int *cut, int & L,
  c_int *s, c_int *e, int *parity, integer depth=0)
{
  // write best minimal cut and return its cut_quality;
  // cerr<<"look for cut "<<endl;
  static integer qu,ne,nv,X,Y,*tcut;

  // TRCA(cut,depth);
  // TRC(qu);
  if(depth==MAX_CUT_SIZE) { ret 0; }
  // cuts of >5 edges are not useful

  if(depth==0) {
    tcut=0;
    qu=1<<30;
    X=g.X;
    Y=g.Y;
    ne=(3*X+Y)/2;
    nv=X+Y;
    ERROR(nv%2);
  }

  // TRC(3);
  integer parity2[nv], cq=-1;

  // do not try <=3 cuts in cubic graphs
  // when we input by default cycl 4-connected ones
  if(Y || depth>3)
  {
    cq=cut_quality(g,cut,depth,qu,s,e,parity2+Y);
    // TRC(4);
    if(cq==0) ret 0;
  }
  // too many unival vertices - extending cut will 
  // not make anything better
  // if at depth 0 shows that cut is not good to use

  if(cq==-1) {
    // not a cut yet - extend
    integer start=(depth ? cut[depth-1]+1 : 0), & w=cut[depth];

    if(Y)
    { for(w=start;w<ne;++w)
      if(s[w]>=0 && e[w]>=0) look_for_cut(g,cut,L,s,e,parity,depth+1); }
    else
    for(w=start;w<ne;++w)
      look_for_cut(g,cut,L,s,e,parity,depth+1);

    // after all is checked 
    if(depth==0) {
      if(tcut!=0) {
        copyarray(tcut,L,cut);
        delete [] tcut;
      }
      else ret 0;
    }
    ret qu;
  }

  ERROR(depth==0);
  //at depth 0 must be able to find cut

  // here we have a cut
  // if cut is better
  if(cq<qu)
  {
    // TRC(cq);
    L=depth;
    copyarray(parity2,nv,parity);
    delete [] tcut;
    tcut=copyarray(cut,L);
    qu=cq;
  }

  ret qu;

}

void split(const graph &g, c_int *cut, c_int L, c_int *s, c_int *e,
  c_int *parity, graph ct[2], int *TT=0, int *UR=0)
  // split graph g along a cut cut[] of L edges into uni3val gr
  // ct[0] and ct[1]; s[], e[] and parity[] are params of
  // g computed in cut_comp()
  // parity[] IS REQUIRED TO BE ACCESSIBLE ALSO WITH INDICES -1 to -g.Y
  // TO WORK FOR UNIVAL VERTICES
  // the first N unival vertices of ct[0], ct[1] are those produced
  // by cut
{
  // ERROR(g.Y)
  int X=g.X, Y=g.Y, i,j, t[X], u[X], cn[X], tc=0, uc=0;
  // t[i]=#of vertex i in g in ct[0] if in ct[0] and -1 else
  // u[i]=#of vertex i in g in ct[1] if in ct[1] and -1 else
  // cn[i]=-1-#of edge in cut of which i is endpt if in such edge 
  // (and in separated graphs this becomes unival vertex labelled
  // cn[i]) and 0 else
  // tc, uc count number of 3val vertices by parity found so far
  // TRC(g);
  // TRCA(s,(3*X+Y)/2);
  // TRCA(e,(3*X+Y)/2);
  // TRCA(cut,L);
  ERROR((X+Y)%2);
  // TRCA(parity-Y,X+Y);
  integer dup_parity[X+Y];
  // TRC(cut_quality(g,cut,L,s,e,dup_parity+Y));
  // TRCA(dup_parity, X+Y);
  // TRCA(parity, X+Y);
  

  for(i=0;i<X;++i) 
  {
    if(parity[i]==1) { t[i]=tc++; u[i]=-1; }
    else             { u[i]=uc++; t[i]=-1; }
    cn[i]=0; // zero out cn along the way
  }

  // the re#ing of the univalent vertices
  // start with -L-1 to create space for newly cut unival vertices
  int _tu[Y], *tu=_tu+Y, t1c=-L, u1c=-L;
  for(i=-1;i>=-Y;--i) 
  {
    if(parity[i]==1) { TT[-t1c-L]=-i; tu[i]=--t1c; }
    else             { UR[-u1c-L]=-i; tu[i]=--u1c; }
  }

  // for(i=0;i<L;++i) {
    // TRC(s[cut[i]]);
    // TRC(e[cut[i]]);
  // }
  for(i=0;i<L;++i) cn[s[cut[i]]]=cn[e[cut[i]]]=-1-i;
  // enter verts of cut edges into cn[]

  int g1[3*tc], g2[3*uc];

  for(i=0;i<X;++i) {
    ERROR(parity[i]-1 && parity[i]-2);
    if(parity[i]==1) {
      for(j=0;j<3;++j)
      {
	c_int &x=g.edges[3*i+j];
	int &y=g1[3*t[i]+j];
        if(x<0) y=tu[x]; else 
	// a univalent vertex is shifted up by the cut size
	// note that a unival vertex adj to a 3-val vartex
	// with parity=n also has parity=n (we do not cut
	// along leaves)
	if((y=t[x])<0) y=cn[i];
	// if first assignment is negative, then neighbored 
	// vertex to i in G is on the other side of cut,
	// thus must find number of cut edge from i in cut library ct[]
      }
    }
    else {
      for(j=0;j<3;++j)
      {
        c_int &x=g.edges[3*i+j];
	int &y=g2[3*u[i]+j];
	if(x<0) y=tu[x]; else if((y=u[x])<0) y=cn[i];
      }
    }

  }
  // TRCA(g1,3*tc);
  // TRCA(g2,3*uc);

  ct[0]=graph(tc,-t1c,g1);
  ct[1]=graph(uc,-u1c,g2);
  // TRC(ct[0]);
  // TRC(ct[1]);
  ERROR((ct[0].X+ct[0].Y)%2);
  ERROR((ct[1].X+ct[1].Y)%2);
}

void connect(const graph &g, poly1var *wsp, bool hard_flag=false);


bool cut_and_connect(const graph &g, poly1var *wsp, c_int *s,
  c_int *e, twoar & edgenr, twoar p[2])
{
  // THIS HAS TO BE REWRITTEN COMPLETELY
  // FORGETTING ANYTH ABOUT REGION STUFF

  // try to find good cut and calculation W_N via
  // decomposing along that cut. ret whether good cut found
  // edgenr and p must be const but the [] operation of array 
  // requires to drop `const'

  // A FUNDAMENTAL ASSUMPTION FOR UNIVAL VERTS TO WORK IS
  // THAT ALL UNIVAL VERTICES POINT INTO THE SAME (SAY,
  // THE INFINITE) REGION!!

  // BUT WE CANNOT ASSUME MORE THAN 1-CONNECTEDNESS
  // there is no chance to avoid no 2-connected graphs
  // THUS MUST DEAL W/ REGIONS BOUNDING THEMSELVES
  // AND PAIRS OF REGS BOUNDED BY 2 EDGES

  // here must copy and modify the start of cut_comp()
  // where I seek cuts

  int X=g.X, Y=g.Y, X1=(X-Y)/2+2, TT[Y], UR[Y], ne=(3*X+Y)/2,
    // 5 must be chenged to 1+maximal sought cut size
    csize, i, j, k, l, m, n, r, z;

  ERROR(Y==0)

  if(X1<16) ret false;
  // cut seeking in graphs with few cycles is not interesting
  // and time consuming

  // TWOAR(reg,X+Y) 
  // #define REG(a,b) reg[(a)+Y][(b)+Y]
  // THIS SHOULD NOW NOT BE USED

  // the region left when going along an edge from vert a to vert b

  // TWOAR(opp,X1) 
  // opp[a][b] will give number of ONE edge between regs a and b
  // if a & b are neighbored
  // while opp[][] is filled, will discover 2-cuts
  // THIS SHOULD NOW NOT BE USED

  int ncut[MAX_CUT_SIZE], np[X+Y];

// HERE MUST WRITE CORRECT CUT SEARCH

  r= look_for_cut(g,ncut,csize,s,e,np);

  if(r<=0) ret false;
  ERROR(csize>MAX_CUT_SIZE);

  // cut is good enough
  // TRC(g);
  // TRCA(ncut,csize);
  // TRC(c);
  graph h[2];
  split(g,ncut,csize,s,e,np+Y,h,TT,UR);
  // TRCA(TT,h[0].Y-csize);
  // TRCA(UR,h[1].Y-csize);
  // TRC(h[0].X);
  // TRC(h[1].X);
  poly1var K[fac(h[0].Y)],H[fac(h[1].Y)];
  connect(h[0],K);
  connect(h[1],H);
  // TRC(h[0]);
  partial_combine(K,H,h[0].Y,h[1].Y,csize,wsp,TT,UR);
  // TRC(h[0]);
  ret true;

}

void connect(const graph &g, poly1var *wsp, bool hard_flag)
{
  // MUST MODIFY connect() to seek good cuts in unitri graphs
  // and decompose along these cuts using

  // MUST ASCERTAIN THAT WORKS FOR 1-CONNECT 13VALENT GRAPHS

  // wsp=poly1var[fac(g.Y)]
  int X=g.X, Y=g.Y, ne=(3*X+Y)/2,i,j,k,p,q,r,ori[X],a[2*ne],
    X1=(X-Y)/2+(Y?2:3), s[ne], e[ne], dv,dv0,sv,sv0,sg=1, perm[Y],
    sg0=Y%2?-1:1, f=fac(Y), cf[f][X1];
    // X1=1+the nax number of closed loops that can occur
    // for a marking (when no unival edges, then the all +
    // marking has an extra component going along the infinite
    // region).
    // cf output polynomial coefficient array
    // perm,rp permutatin and reverse perm for strands with
    // output
    // s[], and e[] wll store start and end of edges
  ERROR((X+Y)%2);

  // ne= # of edges
  // sg = sign of marking
  // 
  // ori[i] = sign of vrt i (0 pos, 1 neg)
  // 
  // a[i] is 0 or >0 dep on whether edge i is passed by some loop
  // counted so far from smaller to larger vertex # (pos direction);
  // this is used to test whether all loops have been found; note that
  // each loop passes an edge in pos direction; so we can forget about
  // counting edges poassed in neg direction
  //
  // s[i],e[i] are start and end vertex of edge i, s.t. s[i]<e[i]
  // edgenr[i][j]=k if k-th edge connects verts i and j (with i<j),
  // o.e. EN(s[i],e[i])==i

  // can work only if this poly1var constructor does not reduce
  // and the = operation of poly1var does not reduce

  // TRC(ne);

  for(i=q=0;i<X;++i) for(j=0;j<3;++j)
  {
    r=0;
    if(q==ne) break;
    s[q]=i; e[q]=g.edges[i*3+j];
    if(e[q]<0) { swap(s[q],e[q]); r=1; } 
    if(r || s[q]<e[q]) {
      // TRC(s[q]);
      // TRC(e[q]);
      ++q;
    }
    // collect edge information: start and end vertex
    // edges between 3val vertices are counted twice,
    // so skip one count
  }

  ERROR(q-ne);

  // need to move edges to unival verts to start
  // of edge list (see below why)
  for(i=0;i<ne;++i) if(s[i]>=0) {
    for(j=i+1;j<ne;++j) if(s[j]<0)
    {
      swap(s[i],s[j]);
      swap(e[i],e[j]); break;
    }
    if(j==ne) break;
  }

  TWOAR(edgenr,X+Y);
#define EN(x,y) edgenr[(x)+Y][(y)+Y]

  // TRCA(s,ne);
  // TRCA(e,ne);
  // set EN: EN(x,y) is the number of edge between verts x & y
  // or -1 if no edge
  // for(i=-Y;i<X;++i) for(j=i+1;j<X;++j) EN(i,j)=EN(j,i)=-1;
  for(i=0;i<ne;++i) EN(s[i],e[i])=EN(e[i],s[i])=i;

  // need here the external edges first since we cannot 
  // close a strand hanging on some of them. and so some
  // strand will be detected only partially if started at 
  // internal edge

  trear pn(2,twoar(X,onear(X+Y)));

  for(i=0;i<X;++i) for(j=0;j<3;++j) for(k=0;k<2;++k)
  pn[k][i][g.edges[3*i+j]+Y]=g.edges[3*i+(j+k+1)%3];
  // pn[k][i][j] is the vertex previous (k=1) or next (p=0)
  // of vertex j among the cyclically ordered vertices adj
  // to 3val vertex i

  if(!hard_flag && X>16 && cut_and_connect(g,wsp,s,e,edgenr,pn)) ret;

  zeroarray(ori,X);
  for(i=0;i<f;++i) zeroarray(cf[i],X1);

  integer inve[f];
  inverse_table(inve,Y);
  // TRCA(inve,f);

  do {
    // TRC(j);
    zeroarray(a,ne,-1);

    for(p=0;k=member(-1,a,ne);++p)
    {
      // TRC(k);

      // p=# of loops (current included)
      // follow a loop starting at edge k-1

      dv0=dv=e[--k]; sv0=sv=s[k];
      // sv0=starting source vertex; dv0=starting dest vertex
      // sv=current source vertex; dv=current dest vertex 

      do {

	ERROR(EN(sv,dv)<0)
	// TRC(X);
	// TRC(Y);
	// TRC(EN(sv,dv));
	// TRC(sv0);
	// TRC(dv0);
	// TRC(sv);
	// TRC(dv);
        if(dv>sv) a[EN(sv,dv)]=p;
	// else a[ne+EN(sv,dv)]=p;
	// mark pos dir edge

	// step to next edge
	if(dv<0) break;
	k=Y+sv;
	dv=pn[ori[dv]][sv=dv][k];
      } while(dv0-dv || sv0-sv);
      // until back @start

      // TRC(k);
      ERROR(sv0>0 && dv<0)
      // ERROR(sv0<0 && dv==sv0)

      if(dv<0) // external strand unclosed
      { ERROR(sv<-Y)
        perm[-1-sv0]=-dv;
	// rp[-1-dv]=-sv0;
	--p;
	// record reverse perm for opposite marking
      }

    } //for(p=0;...)

    int c=code(perm,Y),ci=inve[c];
    // TRC(c);
    cf[c][p]+=sg;      // add mon to W + one of compl marking
    cf[ci][p]+=sg*sg0; // add mon to W + one of compl marking
    // cf[code(perm,Y)][p]+=sg; // add mon to W
    // cf[code(rp  ,Y)][p]+=sg*sg0;
    
    for(j=0;ori[j];ori[j++]=0); ori[j]=1;
    if(++j%2) sg*=-1;

    // TRCA(ori,X);
    // {for(int i=0;i<f;++i) TRCA(cf[i],X1);}
  } while(j<X);

  for(i=0;i<f;++i) wsp[i]=poly1var(0,X1-1,cf[i]);

}

// this is the main work procedure which seeks a 4cut or 5cut and
// calculation W_N after decomp along 4 or 5 cut

poly1var cut_comp(const graph & g)
{
  // static poly1var th("1 3 -2 0 2");
  ERROR(g.Y);
  // TRC(g);

  int X=g.X,ne=3*X/2,q=0,s[ne],e[ne],i,j,k,l,m,n,r;

  TWOAR(nd,X)
#define ND(x,y) nd[x][y]
  // ND(x,y) is the number of the edge between verts x and y
  // TRC(nd[X-1][X-1]);

  for(i=0;i<X;++i) for(j=0;j<3;++j)
  {
    if(q==ne) break;
    s[q]=i;
    e[q]=g.edges[i*3+j];
    if(ND(s[q],e[q])<0) { ND(s[q],e[q])=ND(e[q],s[q])=q; ++q; }
  }
  ERROR(q-ne);

  int a[ne],p, np[X],ncut[MAX_CUT_SIZE],csize; // #regs=#vert/2+2
  zeroarray(a,ne);
  // TWOAR(reg,ne);
  // reg[i][j] is the reg left pf edge
  // between verts i & j when passed in dir from i to j
  // thus reg[i][j]!=reg[j][i]
  // NO REGS ALLOWED

// HERE MUST WRITE CORRECT CUT SEARCH
  // TRC(1);
  r= look_for_cut(g,ncut,csize,s,e,np);
  // TRC(2);

  if(r<=0)
  // in non-planar graphs it manot find good cuts!
  {
    // TRC(3);
    poly1var K[1];
    connect(g,K,true);
    ret K[0];
  }
  // for full grpah there must be an ok cut!
  ERROR(csize>MAX_CUT_SIZE);

  // 4-cut is good enough
  // TRC(g);
  // TRC(c);
  // found 4-cut
  // TRCA(ocut,4);
  int w=fac(csize);
  graph h[2];
  split(g,ncut,csize,s,e,np,h);
  // cout<<"*";
  // cout.flush();
  // TRC(h[0]);
  // TRC(h[1]);
  poly1var K[w],H[w];
  connect(h[0],K);
  connect(h[1],H);
  // TRCA(K,24);
  // TRCA(H,24);
  ret combine(H,K,csize);
} 


#define OK(d) (argc>(d) && argv[d]!=0 && *(argv[d])!='-')

main(int argc, char **argv) {

  graph g; integer number=0; // , ns=0;
  // current graph # and number of initial graphs to skip

  ifstream Y;
  if(OK(1)) Y.open(argv[1],ios::in);

  istream & gr_file= OK(1) ? Y : cin;
  
  // if(OK(2)) sscanf(argv[2],"%d",&ns);

  while(gr_file>>number>>g) {

    ERROR(g.Y);
    // TRC(g);
    cout<<g.X<<" "<<number<<" "<<cut_comp(g)<<endl;
    // GRAPH NR INSERTED 
    // TRCA(t,fac(g.Y));
  }
  cout<<"ended at "<<g.X<<" "<<number<<endl;
}


