// overflow checks & gcd reduction added Mar 12 12
#include<iostream>
#include<fstream>
#include<stdio.h>
#include"mydefs.h"

#define ERROR(a) if(a) { cerr<<"error line "<<__LINE__<<endl; }


//class Rat;
//Rat undef,infty;

void lcm_gcd(int z, int n, u_int &lcm, u_int &gcd)
{
  if(n==0) { gcd=lcm=z; ret; }
  if(z==0) { gcd=lcm=n; ret; }
  if(n<0) n=-n;
  if(z<0) z=-z;
  lcm=1, gcd=1;

  // TRC(z); TRC(n);

  for(int j=1;j<=max(z,n);j+=2) {
    int i=(j==1?2:j);
    bool A=z%i==0, B=n%i==0;
    if(A||B) {
      if(A && B) { gcd*=i; lcm*=i; z/=i,n/=i; }
      else if(A) { lcm*=i; z/=i; }
      else { lcm*=i; n/=i; }
      j-=2;
    }
  }
  // TRC(lcm); TRC(gcd);
}

class Rat {
public:
   int z;
   u_int n;

   Rat(int a=0,u_int b=1): z(a), n(b)
   { 
     if(n==0 && z) z=1; else if(n && z==0) n=1; else if(n&&z) reduce();
   }

   Rat(const char *s)
   { Rat r;

     switch(sscanf(s,"%d%*[ /]%u",&(r.z),&(r.n)))
     { case 0:
	 cerr<<"no correct string `"<<s<<"' "<<endl;
	 r=1;
       CASE 1:
	 r.n=1;
     }

     *this=r;
   }


   Rat operator +(const Rat& r) const { 
     u_int l,g;
     lcm_gcd(r.n,n,l,g);
     ERROR(l<0);
     // TRC(*this);
     // TRC(r);
     // TRC(l);
     // TRC(g);
     int nz=z*(r.n/(int)g)+r.z*(n/(int)g); u_int nn=l;
     // TRC(nz);
     // TRC(nn);
     
     Rat res(nz,nn);
     if(abs(double(*this)+double(r)-double(res))>0.000001)
     {
       cerr<<"!! warning: possible overflow in rational addition"<<endl;
       TRC(*this);
       TRC(r);
       TRC(res);
     }

     ret res;
   }
   Rat operator -() const { return Rat(-z,n); }
   Rat operator -(const Rat& r) const { return *this+(-r); }

   Rat operator *(const Rat& r) const {
     u_int l,g,l1,g1;
     lcm_gcd(r.z,n,l,g);
     lcm_gcd(r.n,z,l1,g1);
     ERROR(g1<0 || g<0);
     int nz=(z/(int)g1)*(r.z/(int)g); u_int nn=(r.n/g1)*(n/g);
     // TRC(z);
     // TRC(g1);
     // TRC(z/(int)g1);
     // TRC(r.z/(int)g);
     // TRC(*this);
     // TRC(r);
     // TRC(nz);
     // TRC(nn);
     // TRC(g1);
     // TRC(g);
     
     Rat res(nz,nn);
     if(abs(double(*this)*double(r)/double(res)-1)>0.000001)
     {
       cerr<<"!! warning: possible overflow in rational mult"<<endl;
       TRC(*this);
       TRC(r);
       TRC(res);
     }
     ret res;
   }

   Rat operator ~() const { 
     // the mult inverse
     if(z==0) {
       cerr<<"error: division by zero; ret undef"<<endl;
       ret Rat(1,0);
     }
     ret z>0? Rat((int)n,(u_int)z) : Rat((int)(-n),(u_int)(-z));
   }
   Rat operator /(const Rat& r) const { 
	return *this * (~r);
   }
   Rat& operator *=(const Rat& r) { return *this=*this*r; }
   Rat& operator /=(const Rat& r) { return *this=*this/r; }
   Rat& operator +=(const Rat& r) { return *this=*this+r; }
   Rat& operator -=(const Rat& r) { return *this=*this-r; }
   Rat& operator ++() { return *this+=1; }
   Rat& operator --() { return *this-=1; }
   Rat operator ++(int) { Rat a=*this; ++*this; return a; }
   Rat operator --(int) { Rat a=*this; --*this; return a; }

   bool operator == (const Rat& r) const { 
     if (n&&r.n) return z*r.n==n*r.z;
     // *this & r regular ratios

     if (n||r.n) return false;
     // one regular, one not

     return r.z==z; 
     // both not regular
   }

   operator double() const { return ((double) z)/n; }
   operator int() const { return z/(int)n; }

   bool operator!=(const Rat& r) const { return !(*this==r); }
   oper bool () const { return *this!=Rat(0); }
   bool operator>(const Rat& r) const { return (double)(*this) > (double)r; }
   bool operator<(const Rat& r) const { return (double)(*this) < (double)r; }
   bool operator>=(const Rat& r) const { return (*this)>r || *this==r; }
   bool operator<=(const Rat& r) const { return (*this)<r || *this==r; }


   friend ostream & operator<<(ostream & o,const Rat& r);

   friend istream & operator>>(istream & i,Rat& r)
   {
     char f[10];
     cin>>f;
     r=f;
     ret i;
   }

   Rat reduce();

} undef(0,0),infty(1,0);

Rat Rat::reduce()
{ 
  u_int l,g; 
  lcm_gcd(z,n,l,g);
  ERROR(n && g<0 || g==0)
  // TRC(g);
  z/=(int)g; n/=g;

  return *this;
}

ostream & operator<<(ostream & o,const Rat & r)
{
  // r==Rat(1,0);
  if(r==undef) return o<<"undef ";
  if(r==infty) return o<<"infty ";
  o<<" "<<r.z;
  return r.n>1 ? o<<"/"<<r.n<<" " : o<<" "; 
}

typedef Rat Rational;
typedef Rat rat;



